{*********************************************************}
{                                                         }
{                 Zeos Database Objects                   }
{         Sybase SQL Anywhere Connectivity Classes        }
{                                                         }
{            Originally written by EgonHugeist            }
{                                                         }
{*********************************************************}

{@********************************************************}
{    Copyright (c) 1999-2012 Zeos Development Group       }
{                                                         }
{ License Agreement:                                      }
{                                                         }
{ This library is distributed in the hope that it will be }
{ useful, but WITHOUT ANY WARRANTY; without even the      }
{ implied warranty of MERCHANTABILITY or FITNESS FOR      }
{ A PARTICULAR PURPOSE.  See the GNU Lesser General       }
{ Public License for more details.                        }
{                                                         }
{ The source code of the ZEOS Libraries and packages are  }
{ distributed under the Library GNU General Public        }
{ License (see the file COPYING / COPYING.ZEOS)           }
{ with the following  modification:                       }
{ As a special exception, the copyright holders of this   }
{ library give you permission to link this library with   }
{ independent modules to produce an executable,           }
{ regardless of the license terms of these independent    }
{ modules, and to copy and distribute the resulting       }
{ executable under terms of your choice, provided that    }
{ you also meet, for each linked independent module,      }
{ the terms and conditions of the license of that module. }
{ An independent module is a module which is not derived  }
{ from or based on this library. If you modify this       }
{ library, you may extend this exception to your version  }
{ of the library, but you are not obligated to do so.     }
{ If you do not wish to do so, delete this exception      }
{ statement from your version.                            }
{                                                         }
{                                                         }
{ The project web site is located on:                     }
{   http://zeos.firmos.at  (FORUM)                        }
{   http://sourceforge.net/p/zeoslib/tickets/ (BUGTRACKER)}
{   svn://svn.code.sf.net/p/zeoslib/code-0/trunk (SVN)    }
{                                                         }
{   http://www.sourceforge.net/projects/zeoslib.          }
{                                                         }
{                                                         }
{                                 Zeos Development Group. }
{********************************************************@}

unit ZDbcSQLAnywhereStatement;

{$I ZDbc.inc}

interface

{$IFNDEF ZEOS_DISABLE_ASA}
uses Classes, {$IFDEF MSEgui}mclasses,{$ENDIF} SysUtils, FmtBCD,
  ZDbcIntfs, ZDbcStatement, ZCompatibility, ZDbcLogging, ZVariant, ZClasses,
  ZDbcSQLAnywhere, ZPlainSQLAnywhere, ZCollections;

type
  {** Implements Prepared SQL Statement. }
  TZAbstractSQLAnywhereStatement = class(TZRawPreparedStatement)
  private
    FSQLAnyConnection: IZSQLAnywhereConnection;
    FPlainDriver: TZSQLAnywherePlainDriver;
    FMoreResults: Boolean;
    FHasOutParams: Boolean;
    Fa_sqlany_stmt: Pa_sqlany_stmt;
    FCallResultCache: TZCollection;
    Fapi_version: Tsacapi_u32;
    procedure ClearCallResultCache;
  private
    function CreateResultSet: IZResultSet;
  public
    constructor Create(const Connection: IZConnection; const SQL: string; Info: TStrings);
    destructor Destroy; override;

    procedure Prepare; override;
    procedure Unprepare; override;

    procedure AfterClose; override;
    function GetMoreResults: Boolean; override;

    function ExecuteQueryPrepared: IZResultSet; override;
    function ExecuteUpdatePrepared: Integer; override;
    function ExecutePrepared: Boolean; override;
  end;

  TZSQLAnywhereStatement = Class(TZAbstractSQLAnywhereStatement)
  public
    constructor Create(const Connection: IZConnection; Info: TStrings);
  End;

  TZSQLAnywherePreparedStatement = class(TZAbstractSQLAnywhereStatement, IZPreparedStatement)
  private
    Fa_sqlany_bind_paramArray: Pa_sqlany_bind_paramArray;
    FIsNullArray: Psacapi_i32Array;
    FLengthArray: Psize_tArray;
    FBindParamSize: NativeInt;
    FParamsDescribed: Boolean; //SQLAynwhere just describes SP's by now those params we do not change, they might be bidirectional
    FBindAgain: Boolean;
  protected
    procedure BindRawStr(Index: Integer; const Value: RawByteString); override;
    procedure BindRawStr(Index: Integer; Buf: PAnsiChar; Len: LengthInt); override;
    procedure BindLob(Index: Integer; SQLType: TZSQLType; const Value: IZBlob); override;
    function InitDataValue(Index: Integer; SQLType: TZSQLType; Length: Tsize_t): Pa_sqlany_data_value;
  protected
    procedure PrepareInParameters; override;
    procedure UnPrepareInParameters; override;
    procedure BindInParameters; override;
    procedure AddParamLogValue(ParamIndex: Integer; SQLWriter: TZRawSQLStringWriter; Var Result: RawByteString); override;
    procedure SetBindCapacity(Capacity: Integer); override;
    procedure CheckParameterIndex(var Value: Integer); override;
  public
    procedure AfterConstruction; override;
  public
    procedure SetNull(Index: Integer; SQLType: TZSQLType);
    procedure SetBoolean(Index: Integer; Value: Boolean);
    procedure SetByte(Index: Integer; Value: Byte);
    procedure SetShort(Index: Integer; Value: ShortInt);
    procedure SetWord(Index: Integer; Value: Word);
    procedure SetSmall(Index: Integer; Value: SmallInt);
    procedure SetUInt(Index: Integer; Value: Cardinal);
    procedure SetInt(Index: Integer; Value: Integer);
    procedure SetULong(Index: Integer; const Value: UInt64);
    procedure SetLong(Index: Integer; const Value: Int64);
    procedure SetFloat(Index: Integer; Value: Single);
    procedure SetDouble(Index: Integer; const Value: Double);
    procedure SetCurrency(Index: Integer; const Value: Currency);
    procedure SetBigDecimal(Index: Integer; const Value: TBCD);
    procedure SetBytes(Index: Integer; const Value: TBytes); reintroduce; overload;
    procedure SetBytes(Index: Integer; Value: PByte; Len: NativeUInt); reintroduce; overload;
    procedure SetGuid(Index: Integer; const Value: TGUID); reintroduce;
    procedure SetDate(Index: Integer; const Value: TZDate); reintroduce; overload;
    procedure SetTime(Index: Integer; const Value: TZTime); reintroduce; overload;
    procedure SetTimestamp(Index: Integer; const Value: TZTimeStamp); reintroduce; overload;
  end;

  TZSQLAnywhereCallableStatement = class(TZAbstractCallableStatement_A, IZCallableStatement)
  protected
    function CreateExecutionStatement(const StoredProcName: String): TZAbstractPreparedStatement; override;
  end;

{$ENDIF ZEOS_DISABLE_ASA}
implementation
{$IFNDEF ZEOS_DISABLE_ASA}

uses ZSysUtils, ZDbcUtils, ZMessages, ZDbcSQLAnywhereResultSet,
  ZDbcGenericResolver, ZEncoding, ZDbcProperties, ZFastCode;

{ TZAbstractSQLAnywhereStatement }

{**
  Constructs this object and assignes the main properties.
  @param Connection a database connection object.
  @param SQL the query
  @param Info a statement parameters.
}
procedure TZAbstractSQLAnywhereStatement.ClearCallResultCache;
var I: Integer;
  RS: IZResultSet;
begin
  for I := 0 to FCallResultCache.Count -1 do
    if Supports(FCallResultCache[i], IZResultSet, RS) then
      RS.Close;
  FreeAndNil(FCallResultCache);
end;

constructor TZAbstractSQLAnywhereStatement.Create(const Connection: IZConnection;
  const SQL: string; Info: TStrings);
begin
  inherited Create(Connection, SQL, Info);

  FSQLAnyConnection := Connection as IZSQLAnywhereConnection;
  FPlainDriver := FSQLAnyConnection.GetPlainDriver;
  ResultSetType := rtScrollSensitive;
  Fapi_version := FSQLAnyConnection.Get_api_version;
end;

destructor TZAbstractSQLAnywhereStatement.Destroy;
begin
  inherited Destroy;
  FSQLAnyConnection := nil;
end;

function TZAbstractSQLAnywhereStatement.CreateResultSet: IZResultSet;
var
  NativeResultSet: TZSQLAnywhereResultSet;
  CachedResultSet: TZSQLAnywhereCachedResultSet;
begin
  With FSQLAnyConnection do begin
    NativeResultSet := TZSQLAnywhereResultSet.Create(Self, SQL, @Fa_sqlany_stmt);
    if ResultSetConcurrency = rcUpdatable then begin
      CachedResultSet := TZSQLAnywhereCachedResultSet.Create(NativeResultSet, SQL, nil, ConSettings);
      CachedResultSet.SetResolver(TZGenerateSQLCachedResolver.Create(Self, NativeResultSet.GetMetadata));
      CachedResultSet.SetConcurrency(GetResultSetConcurrency);
      Result := CachedResultSet;
    end else
      Result := NativeResultSet;
    FOpenResultSet := Pointer(Result);
  end;
end;

procedure TZAbstractSQLAnywhereStatement.Prepare;
begin
  if FCallResultCache <> nil then
    ClearCallResultCache;
  if not Prepared then begin
    Fa_sqlany_stmt := FplainDriver.sqlany_prepare(FSQLAnyConnection.Get_a_sqlany_connection,
      Pointer(fASQL));
    if Fa_sqlany_stmt = nil then
      FSQLAnyConnection.HandleError(lcPrepStmt, fASQL, Self);
    inherited Prepare;
  end;
end;

procedure TZAbstractSQLAnywhereStatement.Unprepare;
begin
  if FCallResultCache <> nil then
    ClearCallResultCache;
  inherited Unprepare;
  if Fa_sqlany_stmt <> nil then begin
    FPLainDriver.sqlany_free_stmt(Fa_sqlany_stmt);
    Fa_sqlany_stmt := nil;
  end;
end;

procedure TZAbstractSQLAnywhereStatement.AfterClose;
begin
end;

{**
  Moves to a <code>Statement</code> object's next result.  It returns
  <code>true</code> if this result is a <code>ResultSet</code> object.
  This method also implicitly closes any current <code>ResultSet</code>
  object obtained with the method <code>getResultSet</code>.

  <P>There are no more results when the following is true:
  <PRE>
        <code>(!getMoreResults() && (getUpdateCount() == -1)</code>
  </PRE>

 @return <code>true</code> if the next result is a <code>ResultSet</code> object;
   <code>false</code> if it is an update count or there are no more results
 @see #execute
}
function TZAbstractSQLAnywhereStatement.GetMoreResults: Boolean;
begin
  Result := FPlainDriver.sqlany_get_next_result(Fa_sqlany_stmt) = 1;
  if Result then begin
  end;
end;

{**
  Executes an SQL statement that may return multiple results.
  Under some (uncommon) situations a single SQL statement may return
  multiple result sets and/or update counts.  Normally you can ignore
  this unless you are (1) executing a stored procedure that you know may
  return multiple results or (2) you are dynamically executing an
  unknown SQL string.  The  methods <code>execute</code>,
  <code>getMoreResults</code>, <code>getResultSet</code>,
  and <code>getUpdateCount</code> let you navigate through multiple results.

  The <code>execute</code> method executes an SQL statement and indicates the
  form of the first result.  You can then use the methods
  <code>getResultSet</code> or <code>getUpdateCount</code>
  to retrieve the result, and <code>getMoreResults</code> to
  move to any subsequent result(s).

  @return <code>true</code> if the next result is a <code>ResultSet</code> object;
  <code>false</code> if it is an update count or there are no more results
  @see #getResultSet
  @see #getUpdateCount
  @see #getMoreResults
}
function TZAbstractSQLAnywhereStatement.ExecutePrepared: Boolean;
var num_cols: Tsacapi_i32;
begin
  PrepareOpenResultSetForReUse;
  Prepare;
  if FWeakIntfPtrOfIPrepStmt <> nil then
    BindInParameters;
  if FMoreResults or FHasOutParams
  then LastResultSet := ExecuteQueryPrepared
  else begin
    if FplainDriver.sqlany_execute(Fa_sqlany_stmt) <> 1 then
      FSQLAnyConnection.HandleError(lcExecute, fASQL, Self);
    num_cols := FplainDriver.sqlany_num_cols(Fa_sqlany_stmt);
    if num_cols < 0
    then FSQLAnyConnection.HandleError(lcExecute, fASQL, Self)
    else if num_cols > 0
    then LastResultSet := CreateResultSet
    else LastUpdateCount := FplainDriver.sqlany_affected_rows(Fa_sqlany_stmt);
  end;
  Result := Assigned(FLastResultSet);
  { Logging SQL Command and values}
  if DriverManager.HasLoggingListener then
    DriverManager.LogMessage(lcExecPrepStmt,Self);
end;

{**
  Executes the SQL query in this <code>PreparedStatement</code> object
  and returns the result set generated by the query.

  @return a <code>ResultSet</code> object that contains the data produced by the
    query; never <code>null</code>
}
function TZAbstractSQLAnywhereStatement.ExecuteQueryPrepared: IZResultSet;
var num_cols: Tsacapi_i32;
begin
  PrepareOpenResultSetForReUse;
  Prepare;
  if FWeakIntfPtrOfIPrepStmt <> nil then
    BindInParameters;
  if FplainDriver.sqlany_execute(Fa_sqlany_stmt) <> 1 then
    FSQLAnyConnection.HandleError(lcExecute, fASQL, Self);
  Result := nil;
  num_cols := FplainDriver.sqlany_num_cols(Fa_sqlany_stmt);
  if num_cols > 0
  then Result := CreateResultSet
  else begin
    while GetMoreResults and (FLastResultSet = nil) do ;
    Result := GetResultSet;
  end;
  { Logging SQL Command and values}
  if DriverManager.HasLoggingListener then
    DriverManager.LogMessage(lcExecPrepStmt,Self);
end;

{**
  Executes the SQL INSERT, UPDATE or DELETE statement
  in this <code>PreparedStatement</code> object.
  In addition,
  SQL statements that return nothing, such as SQL DDL statements,
  can be executed.

  @return either the row count for INSERT, UPDATE or DELETE statements;
  or 0 for SQL statements that return nothing
}
function TZAbstractSQLAnywhereStatement.ExecuteUpdatePrepared: Integer;
var num_cols: Tsacapi_i32;
begin
  Prepare;
  if FWeakIntfPtrOfIPrepStmt <> nil then
    BindInParameters;
  if FplainDriver.sqlany_execute(Fa_sqlany_stmt) <> 1 then
    FSQLAnyConnection.HandleError(lcExecute, fASQL, Self);
  Result := -1;
  num_cols := FplainDriver.sqlany_num_cols(Fa_sqlany_stmt);
  if num_cols = 0
  then Result := FplainDriver.sqlany_affected_rows(Fa_sqlany_stmt)
  else while GetMoreResults and (FLastResultSet <> nil) do
    Result := FplainDriver.sqlany_affected_rows(Fa_sqlany_stmt);
  { Logging SQL Command and values }
  if DriverManager.HasLoggingListener then
    DriverManager.LogMessage(lcExecPrepStmt,Self);
  LastUpdateCount := Result;
end;

{ TZSQLAnywherePreparedStatement }

procedure TZSQLAnywherePreparedStatement.AfterConstruction;
begin
  inherited;
  if Fapi_version >= SQLANY_API_VERSION_4
  then FBindParamSize := SizeOf(Ta_sqlany_bind_paramV4up)
  else FBindParamSize := SizeOf(Ta_sqlany_bind_param)
end;

procedure TZSQLAnywherePreparedStatement.BindInParameters;
var I: Integer;
begin
  if FBindAgain then begin
    for i := 0 to BindList.Count -1 do
      {$R-}
      if FPlainDriver.sqlany_bind_param(Fa_sqlany_stmt, Cardinal(I),
        Pointer(PAnsiChar(Fa_sqlany_bind_paramArray) + (FBindParamSize * i))) <> 1 then
      {$IFDEF RangeCheckEnabled}{$R+}{$ENDIF}
        FSQLAnyConnection.HandleError(lcExecute, 'sqlany_bind_param', Self);
    FBindAgain := False;
  end;
  inherited BindInParameters;
end;

procedure TZSQLAnywherePreparedStatement.BindLob(Index: Integer; SQLType: TZSQLType;
  const Value: IZBlob);
var data_value: Pa_sqlany_data_value;
  Stream: TStream;
begin
  inherited BindLob(Index, SQLType, Value); //else FPC raises tons of memleaks
  data_value := InitDataValue(Index, SQLType, 0);
  if (Value = nil) or Value.IsEmpty
  then data_value.is_null^ := 1
  else begin
    data_value.is_null^ := 0;
    Stream := Value.GetStream;
    try
      data_value.length^ := Stream.Size;
      data_value.buffer_size := ((data_value.length^ shr 3)+1) shl 3;
      ReallocMem(data_value.buffer, data_value.buffer_size);
      Stream.Read(data_value.buffer^, data_value.length^);
    finally
      Stream.Free;
    end;
  end;
end;

procedure TZSQLAnywherePreparedStatement.BindRawStr(Index: Integer;
  const Value: RawByteString);
begin
  if Pointer(Value) <> nil
  then BindRawStr(Index, Pointer(Value), Length(Value))
  else BindRawStr(Index, PEmptyAnsiString, 0);
end;

procedure TZSQLAnywherePreparedStatement.BindRawStr(Index: Integer; Buf: PAnsiChar;
  Len: LengthInt);
var data_value: Pa_sqlany_data_value;
begin
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stString, Len);
  data_value.is_null^ := 0;
  data_value.length^ := Len;
  if Len > 0 then
    Move(Buf^, data_value.buffer^, Len);
end;

procedure TZSQLAnywherePreparedStatement.CheckParameterIndex(
  var Value: Integer);
var I: Integer;
begin
  if not Prepared then
    Prepare;
  if (Value<0) or (Value+1 > BindList.Count) then
    raise EZSQLException.Create(SInvalidInputParameterCount);
  if BindList.HasOutOrInOutOrResultParam then
    for I := 0 to Value do
      if Ord(BindList[I].ParamType) > Ord(pctInOut) then
        Dec(Value);
end;

function TZSQLAnywherePreparedStatement.InitDataValue(Index: Integer;
  SQLType: TZSQLType; Length: Tsize_t): Pa_sqlany_data_value;
label jmpVarLen;
var ActSize: Tsize_t;
    ActType: Ta_sqlany_data_type;
    Bind: Pa_sqlany_bind_param;
begin
  Bind := Pointer(PAnsiChar(Fa_sqlany_bind_paramArray) + (FBindParamSize * Index));
  Result := @Bind.value;
  ActSize := Result.buffer_size;
  ActType := Result._type;
  case SQLType of
    stBoolean, stByte: begin
       Result.buffer_size := SizeOf(Byte);
       Result._type := A_UVAL8;
      end;
    stShort: begin
       Result.buffer_size := SizeOf(ShortInt);
       Result._type := A_VAL8;
      end;
    stWord: begin
       Result.buffer_size := SizeOf(Word);
       Result._type := A_UVAL16;
      end;
    stSmall: begin
       Result.buffer_size := SizeOf(Word);
       Result._type := A_VAL16;
      end;
    stLongWord: begin
       Result.buffer_size := SizeOf(Cardinal);
       Result._type := A_UVAL32;
      end;
    stInteger: begin
       Result.buffer_size := SizeOf(Integer);
       Result._type := A_VAL32;
      end;
    stULong: begin
       Result.buffer_size := SizeOf(Uint64);
       Result._type := A_UVAL64;
      end;
    stLong: begin
       Result.buffer_size := SizeOf(Int64);
       Result._type := A_VAL64;
      end;
    stFloat, stDouble: begin
       Result.buffer_size := SizeOf(Double);
       Result._type := A_DOUBLE;
      end;
    stCurrency,
    stBigDecimal: begin
       Result.buffer_size := 44;
       Result._type := A_STRING;
      end;
    stDate: begin
       Result.buffer_size := 16;
       Result._type := A_STRING;
      end;
    stTime: begin
       Result.buffer_size := 24;
       Result._type := A_STRING;
      end;
    stTimestamp: begin
       Result.buffer_size := 32;
       Result._type := A_STRING;
      end;
    stGUID: begin
       Result.buffer_size := 44;
       Result._type := A_STRING;
      end;
    stString, stUnicodeString: begin
        Result._type := A_STRING;
        goto jmpVarLen;
      end;
    stBytes: begin
        Result._type := A_BINARY;
jmpVarLen:
        if length = 0 then
          length := 8
        else if Result.buffer_size < length then
          length := ((length shr 3)+1) shl 3; //8 Byte aligned incl. space for zero term
        Result.buffer_size := length;
      end;
    stAsciiStream,
    stUnicodeStream: begin
       Result.buffer_size := 0;
       Result._type := A_STRING;
      end;
    stBinaryStream: begin
       Result.buffer_size := 0;
       Result._type := A_BINARY;
      end;
    else raise EZSQLException.Create(SUnsupportedParameterType);
  end;
  if (ActSize <> Result.buffer_size) then begin
    FBindAgain := True;
    ReallocMem(Result.buffer, Result.buffer_size);
  end;
  FBindAgain := FBindAgain or (ActType <> Result._type);
end;

{**
  Prepares eventual structures for binding input parameters.
}
procedure TZSQLAnywherePreparedStatement.PrepareInParameters;
var num_params, I: Tsacapi_i32;
  Bind: Pa_sqlany_bind_param;
begin
  FBindAgain := True;
  inherited PrepareInParameters;
  num_params := FPlainDriver.sqlany_num_params(Fa_sqlany_stmt);
  if num_params = -1 then
    FSQLAnyConnection.HandleError(lcExecute, fASQL, Self);
  SetBindCapacity(num_params);
  for i := 0 to num_params -1 do begin
    {$R-}
    Bind := Pointer(PAnsiChar(Fa_sqlany_bind_paramArray) + (FBindParamSize * I));
    { in V2 just IO is descibed }
    if FPlainDriver.sqlany_describe_bind_param(Fa_sqlany_stmt, I, Bind) <> 1 then
      FSQLAnyConnection.HandleError(lcExecute, fASQL, Self);
    FParamsDescribed := FParamsDescribed or (Bind.value._type <> A_INVALID_TYPE);
    Bind.value.length :=  @FLengthArray[i];
    Bind.value.is_null := @FIsNullArray[i];
    if Bind.value.buffer_size > 0 then
      GetMem(Bind.value.buffer, Bind.value.buffer_size);

    {$IFDEF RangeCheckEnabled}{$R+}{$ENDIF}
  end;
end;

procedure TZSQLAnywherePreparedStatement.SetBigDecimal(Index: Integer;
  const Value: TBCD);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stBigDecimal, 0);
  data_value.is_null^ := 0;
  data_value.length^ := BcdToRaw(Value, data_value.buffer, '.');
end;

procedure TZSQLAnywherePreparedStatement.SetBindCapacity(Capacity: Integer);
var OldCapacity, I: Integer;
  Bind: Pa_sqlany_bind_param;
begin
  OldCapacity := Bindlist.Capacity;
  if (Fa_sqlany_stmt <> nil) and (BindList.Count > 0) then
    for i := OldCapacity -1 downto Capacity do
      FplainDriver.sqlany_bind_param(Fa_sqlany_stmt, Cardinal(I), nil);

  inherited SetBindCapacity(Capacity);
  if OldCapacity <> Capacity then begin
    BindList.SetCount(Capacity);
    ReallocMem(FIsNullArray, Capacity * SizeOf(Tsacapi_i32));
    ReallocMem(FLengthArray, Capacity * SizeOf(Tsize_t));
    for i := OldCapacity-1 downto Capacity do begin
      Bind := Pointer(PAnsiChar(Fa_sqlany_bind_paramArray) + (FBindParamSize * i));
      if Bind^.value.buffer <> nil then
        FreeMem(Bind^.value.buffer);
    end;
    ReallocMem(Fa_sqlany_bind_paramArray, Capacity*FBindParamSize);
    if Capacity > 0 then
      FillChar((PAnsichar(Fa_sqlany_bind_paramArray)+OldCapacity*SizeOf(FBindParamSize))^,
        (Capacity-OldCapacity)*FBindParamSize, #0);
  end;
end;

procedure TZSQLAnywherePreparedStatement.SetBoolean(Index: Integer;
  Value: Boolean);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stBoolean, 0);
  data_value.is_null^ := 0;
  PByte(data_value.buffer)^ := Byte(Value);
end;

procedure TZSQLAnywherePreparedStatement.SetByte(Index: Integer; Value: Byte);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stByte, 0);
  data_value.is_null^ := 0;
  PByte(data_value.buffer)^ := Value;
end;

{**
  Sets the designated parameter to a Java array of bytes by reference.
  The driver converts this to an SQL <code>VARBINARY</code> or
  <code>LONGVARBINARY</code> (depending on the argument's size relative to
  the driver's limits on
  <code>VARBINARY</code> values) when it sends it to the database.

  @param parameterIndex the first parameter is 1, the second is 2, ...
  @param Value the parameter value address
  @param Len the length of the addressed value
}
procedure TZSQLAnywherePreparedStatement.SetBytes(Index: Integer; Value: PByte;
  Len: NativeUInt);
var data_value: Pa_sqlany_data_value;
begin
  if (Len = 0) or (Value = nil) then
    SetNull(Index, stBytes)
  else begin
    {$IFNDEF GENERIC_INDEX}
    Index := Index -1;
    {$ENDIF}
    CheckParameterIndex(Index);
    data_value := InitDataValue(Index, stBytes, Len);
    data_value.is_null^ := 0;
    Move(Value^, data_value.buffer^, Len);
    data_value.length^ := Len;
  end;
end;

{**
  Sets the designated parameter to a Java array of bytes.  The driver converts
  this to an SQL <code>VARBINARY</code> or <code>LONGVARBINARY</code>
  (depending on the argument's size relative to the driver's limits on
  <code>VARBINARY</code> values) when it sends it to the database.

  @param parameterIndex the first parameter is 1, the second is 2, ...
  @param x the parameter value
}
procedure TZSQLAnywherePreparedStatement.SetBytes(Index: Integer;
  const Value: TBytes);
var Len: LengthInt;
begin
  Len := Length(Value);
  if Len = 0
  then SetNull(Index, stBytes)
  else SetBytes(Index, Pointer(Value), Len);
end;

procedure TZSQLAnywherePreparedStatement.SetCurrency(Index: Integer;
  const Value: Currency);
var data_value: Pa_sqlany_data_value;
  P: PAnsiChar;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stCurrency, 0);
  data_value.is_null^ := 0;
  CurrToRaw(Value, data_value.buffer, @P);
  data_value.length^ := P - data_value.buffer;
end;

procedure TZSQLAnywherePreparedStatement.SetDate(Index: Integer;
  const Value: TZDate);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stDate, 0);
  data_value.is_null^ := 0;
  data_value.length^ := ZSysUtils.DateToRaw(Value.Year, Value.Month, Value.Day,
    data_value.buffer, ConSettings.WriteFormatSettings.DateFormat, False, Value.IsNegative)
end;

procedure TZSQLAnywherePreparedStatement.SetDouble(Index: Integer;
  const Value: Double);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stDouble, 0);
  data_value.is_null^ := 0;
  PDouble(data_value.buffer)^ := Value;
end;

procedure TZSQLAnywherePreparedStatement.SetFloat(Index: Integer;
  Value: Single);
begin
  SetDouble(Index, Value);
end;

procedure TZSQLAnywherePreparedStatement.SetGuid(Index: Integer;
  const Value: TGUID);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stGUID, 0);
  data_value.is_null^ := 0;
  ZSysUtils.GUIDToBuffer(@Value.D1, data_value.buffer, [guidWithBrackets]);
  data_value.length^ := 38;
end;

procedure TZSQLAnywherePreparedStatement.AddParamLogValue(ParamIndex: Integer;
  SQLWriter: TZRawSQLStringWriter; var Result: RawByteString);
var Bind: Pa_sqlany_bind_param;
begin
  CheckParameterIndex(ParamIndex);
  Bind := Pointer(PAnsiChar(Fa_sqlany_bind_paramArray) + (FBindParamSize * ParamIndex));
  if (Bind.value.is_null <> nil) and (Bind.value.is_null^ <> 0) then
    SQLWriter.AddText('(NULL)', Result)
  else case Bind.value._type of
    A_BINARY: if BindList[ParamIndex].SQLType = stBinaryStream
              then SQLWriter.AddText('(BLOB)', Result)
              else SQLWriter.AddHexBinary(PByte(Bind.value.buffer), Bind.value.length^, True, Result);
    A_STRING: case BindList[ParamIndex].SQLType of
                stAsciiStream: SQLWriter.AddText('(CLOB)', Result);
                stUnicodeStream: SQLWriter.AddText('(NCLOB)', Result);
                stCurrency, stBigDecimal: SQLWriter.AddText(Bind.value.buffer, Bind.value.length^, Result);
                else SQLWriter.AddTextQuoted(Bind.value.buffer, Bind.value.length^, #39, Result);
              end;
    A_DOUBLE: if Bind.value.buffer_size = SizeOf(Double)
              then SQLWriter.AddFloat(PDouble(Bind.value.buffer)^, Result)
              else SQLWriter.AddFloat(PSingle(Bind.value.buffer)^, Result);
    A_VAL64:  SQLWriter.AddOrd(PInt64(Bind.value.buffer)^, Result);
    A_UVAL64: SQLWriter.AddOrd(PUInt64(Bind.value.buffer)^, Result);
    A_VAL32:  SQLWriter.AddOrd(PInteger(Bind.value.buffer)^, Result);
    A_UVAL32: SQLWriter.AddOrd(PCardinal(Bind.value.buffer)^, Result);
    A_VAL16:  SQLWriter.AddOrd(PSmallInt(Bind.value.buffer)^, Result);
    A_UVAL16: SQLWriter.AddOrd(PWord(Bind.value.buffer)^, Result);
    A_VAL8:   SQLWriter.AddOrd(PShortInt(Bind.value.buffer)^, Result);
    A_UVAL8:  SQLWriter.AddOrd(PByte(Bind.value.buffer)^, Result);
    else SQLWriter.AddText('(UNKNOWN)', Result);
  end;
end;

procedure TZSQLAnywherePreparedStatement.SetInt(Index, Value: Integer);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stInteger, 0);
  data_value.is_null^ := 0;
  PInteger(data_value.buffer)^ := Value;
end;

procedure TZSQLAnywherePreparedStatement.SetLong(Index: Integer;
  const Value: Int64);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stLong, 0);
  data_value.is_null^ := 0;
  PInt64(data_value.buffer)^ := Value;
end;

procedure TZSQLAnywherePreparedStatement.SetNull(Index: Integer;
  SQLType: TZSQLType);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, SQLType, 0);
  data_value.is_null^ := 1;
end;

procedure TZSQLAnywherePreparedStatement.SetShort(Index: Integer;
  Value: ShortInt);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stShort, 0);
  data_value.is_null^ := 0;
  PShortInt(data_value.buffer)^ := Value;
end;

procedure TZSQLAnywherePreparedStatement.SetSmall(Index: Integer;
  Value: SmallInt);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stSmall, 0);
  data_value.is_null^ := 0;
  PSmallInt(data_value.buffer)^ := Value;
end;

procedure TZSQLAnywherePreparedStatement.SetTime(Index: Integer;
  const Value: TZTime);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stTime, 0);
  data_value.is_null^ := 0;
  data_value.length^ := ZSysUtils.TimeToRaw(Value.Hour, Value.Minute, Value.Second,
    Value.Fractions, data_value.buffer, ConSettings.WriteFormatSettings.TimeFormat,
    False, Value.IsNegative)
end;

procedure TZSQLAnywherePreparedStatement.SetTimestamp(Index: Integer;
  const Value: TZTimeStamp);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stTimeStamp, 0);
  data_value.is_null^ := 0;
  data_value.length^ := DateTimeToRaw(Value.Year, Value.Month, Value.Day,
    Value.Hour, Value.Minute, Value.Second, Value.Fractions, data_value.buffer,
    ConSettings.
    WriteFormatSettings.DateTimeFormat, False, Value.IsNegative)
end;

procedure TZSQLAnywherePreparedStatement.SetUInt(Index: Integer;
  Value: Cardinal);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stLongWord, 0);
  data_value.is_null^ := 0;
  PCardinal(data_value.buffer)^ := Value;
end;

procedure TZSQLAnywherePreparedStatement.SetULong(Index: Integer;
  const Value: UInt64);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stULong, 0);
  data_value.is_null^ := 0;
  PUInt64(data_value.buffer)^ := Value;
end;

procedure TZSQLAnywherePreparedStatement.SetWord(Index: Integer; Value: Word);
var data_value: Pa_sqlany_data_value;
begin
  {$IFNDEF GENERIC_INDEX}
  Index := Index -1;
  {$ENDIF}
  CheckParameterIndex(Index);
  data_value := InitDataValue(Index, stWord, 0);
  data_value.is_null^ := 0;
  PWord(data_value.buffer)^ := Value;
end;

{**
  Removes eventual structures for binding input parameters.
}
procedure TZSQLAnywherePreparedStatement.UnPrepareInParameters;
begin
  SetBindCapacity(0);
  FParamsDescribed := False;
  FBindAgain := True;
end;

{ TZSQLAnywhereStatement }

{**
  Constructs this object and assignes the main properties.
  @param Connection a database connection object.
  @param Info a statement parameters.
}
constructor TZSQLAnywhereStatement.Create(const Connection: IZConnection;
  Info: TStrings);
begin
  inherited Create(Connection, '', Info);
end;

{ TZSQLAnywhereCallableStatement }

function TZSQLAnywhereCallableStatement.CreateExecutionStatement(
  const StoredProcName: String): TZAbstractPreparedStatement;
var
  I: Integer;
  SQL: {$IF defined(FPC) and defined(WITH_RAWBYTESTRING)}RawByteString{$ELSE}String{$IFEND};
  SQLWriter: TZSQLStringWriter;
begin
  SQL := '';
  I := Length(StoredProcName);
  i := I + 6+BindList.Count shl 1;
  SQLWriter := TZSQLStringWriter.Create(I);
  SQLWriter.AddText('CALL ', SQL);
  SQLWriter.AddText(StoredProcName, SQL);
  if BindList.Count > 0 then
    SQLWriter.AddChar('(', SQL);
  for i := 0 to BindList.Count-1 do
    if BindList.ParamTypes[i] <> pctReturn then
      SQLWriter.AddText('?,', SQL);
  if BindList.Count > 0 then begin
    SQLWriter.CancelLastComma(SQL);
    SQLWriter.AddChar(')', SQL);
  end;
  SQLWriter.Finalize(SQL);
  FreeAndNil(SQLWriter);
  Result := TZSQLAnywherePreparedStatement.Create(Connection , SQL, Info);
  TZSQLAnywherePreparedStatement(Result).Prepare;
end;

initialization
{$ENDIF ZEOS_DISABLE_ASA}
end.
